class
	AST_FUNCTION_DECLARATION

inherit
	AST_EIFFEL_FEATURE
		rename
			type as return_type
		redefine
			is_extern,
			make_copy
		end
	ADDRESS_INFO

create
	make,
	make_forward_declaration,
	make_from_antlr,
	make_copy

feature {NONE}

	make (a_return_type: AST_TYPE; a_identifier: STRING; a_parameter_list: AST_PARAMETER_LIST)
		require
			a_return_type /= Void and then a_return_type.parent = Void
			a_identifier /= Void and then not a_identifier.is_empty
			a_parameter_list /= Void and then a_parameter_list.parent = Void
		do
			make_forward_declaration (a_return_type, a_identifier, a_parameter_list)
			locals := create {TWO_WAY_LIST [AST_VARIABLE_DECLARATION]}.make
			create statements.make
			statements.parent := Current

			oo_preconditions := create {TWO_WAY_LIST [AST_EXPRESSION]}.make
			oo_postconditions := create {TWO_WAY_LIST [AST_EXPRESSION]}.make
		end

	make_forward_declaration (a_return_type: AST_TYPE; a_identifier: STRING; a_parameter_list: AST_PARAMETER_LIST)
		require
			a_return_type /= Void and then a_return_type.parent = Void
			a_identifier /= Void and then not a_identifier.is_empty
			a_parameter_list /= Void and then a_parameter_list.parent = Void
		do
			storage_class_specifiers := create {TWO_WAY_LIST [INTEGER]}.make
			return_type := a_return_type
			return_type.parent := Current
			identifier := a_identifier
			orig_c_identifier := identifier.twin
			parameter_list := a_parameter_list
			parameter_list.parent := Current

			oo_preconditions := create {TWO_WAY_LIST [AST_EXPRESSION]}.make
			oo_postconditions := create {TWO_WAY_LIST [AST_EXPRESSION]}.make
		end

	make_from_antlr (node: ANTLR_AST_NODE; a_parent: AST_NODE; a_source_info: SOURCE_INFO)
		require
			node /= Void and then node.type = Ast_function_declaration
		local
			cur_head: ANTLR_AST_NODE
			cur_parameter_list: AST_PARAMETER_LIST
		do
			parent := a_parent

			-- storage_class_specifier*
			node.children.start
			from
				storage_class_specifiers := create {TWO_WAY_LIST [INTEGER]}.make
			until
				node.children.item.type /= Ast_storage_class_specifier
			loop
				storage_class_specifiers.extend (node.children.item.children.first.type)
				node.children.forth
			end

			-- attribute_specifier?
			if node.children.item.type = C_kw_attribute then
				update_gcc_attribute_specifier (node.children.item)
				node.children.forth
			end

			-- type (reused below; no parent just yet)
			return_type := process_type (node.children.item, Void)
			node.children.forth

			-- attribute_specifier?
			if node.children.item.type = C_kw_attribute then
				update_gcc_attribute_specifier (node.children.item)
				node.children.forth
			end

			-- function_declaration_head
			from
				cur_head := node.children.item
			until
				cur_head = Void
			loop
				if cur_head.children.first.type /= Ast_pointer_type then
					cur_head.children.start
					if cur_head.children.item.type = C_kw_attribute then
						update_gcc_attribute_specifier (cur_head.children.item)
						cur_head.children.forth
					end
					identifier := cur_head.children.item.token_text
					orig_c_identifier := identifier.twin
					if cur_head.children.last.type = Ast_parameter_list then
						parameter_list := create {AST_PARAMETER_LIST}.make_from_antlr (cur_head.children.last, Current)
					else
						parameter_list := create {AST_PARAMETER_LIST}.make
						parameter_list.parent := Current
					end
					return_type.parent := Current
					cur_head := Void
				else
					if cur_head.children.last.type = Ast_parameter_list then
						cur_parameter_list := create {AST_PARAMETER_LIST}.make_from_antlr (cur_head.children.last, Void)
					else
						cur_parameter_list := create {AST_PARAMETER_LIST}.make
					end
					return_type := create {AST_POINTER_TYPE}.make_from_antlr (
							cur_head.children.first,
							create {AST_FUNCTION_TYPE}.make (return_type, cur_parameter_list),
							Void
						)
					cur_head := cur_head.children.i_th (2)
				end
			end
			node.children.forth

			if not node.children.after then
				if node.children.item.type = C_kw_attribute then
					update_gcc_attribute_specifier (node.children.item)
				else
					-- (variable_declaration* compound_statement)?
					from
						locals := create {TWO_WAY_LIST [AST_VARIABLE_DECLARATION]}.make
					until
						node.children.item.type /= Ast_variable_declaration_or_type
					loop
						locals.extend (create {AST_VARIABLE_DECLARATION}.make_from_antlr (node.children.item, Current, Void))
						node.children.forth
					end
					create statements.make_from_antlr (node.children.item, Current)
				end
			end

			source_info := a_source_info

			oo_preconditions := create {TWO_WAY_LIST [AST_EXPRESSION]}.make
			oo_postconditions := create {TWO_WAY_LIST [AST_EXPRESSION]}.make
		end

feature {ANY} -- data

	return_type: AST_TYPE assign set_return_type

	parameter_list: AST_PARAMETER_LIST

	locals: LINKED_LIST [AST_VARIABLE_DECLARATION]

	statements: AST_COMPOUND_STATEMENT

	set_return_type (a_return_type: AST_TYPE)
		require
			a_return_type /= Void and then a_return_type.parent = Void
		do
			return_type.parent := Void
			return_type := a_return_type
			return_type.parent := Current
		end

feature {ANY}

	oo_preconditions: LIST [AST_EXPRESSION]

	oo_postconditions: LIST [AST_EXPRESSION]

feature {ANY} -- helpers

	is_extern: BOOLEAN
		do
			Result := (storage_class_specifiers.has (C_kw_extern) and locals = Void)
		end

	is_regular: BOOLEAN
		do
			Result := (not is_extern and locals /= Void)
		end

	is_forward_declaration: BOOLEAN
		do
			Result := (not is_extern and locals = Void)
		end

feature {ANY} -- C2Eif data

	cfg: CFG_BASIC_BLOCK assign set_cfg

	set_cfg (a_cfg: CFG_BASIC_BLOCK)
		do
			cfg := a_cfg
		end

	has_jump_statements: BOOLEAN assign set_has_jump_statements

	set_has_jump_statements (a_has_jump_statements: BOOLEAN)
		do
			has_jump_statements := a_has_jump_statements
		end

	return_statements_removed: BOOLEAN assign set_return_statements_removed

	set_return_statements_removed (a_return_statements_removed: BOOLEAN)
		do
			return_statements_removed := a_return_statements_removed
		end

	needs_alloca_factory: BOOLEAN assign set_needs_alloca_factory

	set_needs_alloca_factory (a_needs_alloca_factory: BOOLEAN)
		do
			needs_alloca_factory := a_needs_alloca_factory
		end

	is_longjmp_target: BOOLEAN assign set_is_longjmp_target

	set_is_longjmp_target (a_is_longjmp_target: BOOLEAN)
		do
			is_longjmp_target := a_is_longjmp_target
		end

feature {NONE}

	update_gcc_attribute_specifier (a_node: ANTLR_AST_NODE)
		require
			a_node /= Void and then a_node.type = C_kw_attribute
		do
			if gcc_attribute_specifier = Void then
				create gcc_attribute_specifier.make_from_antlr (a_node, Current)
			else
				gcc_attribute_specifier.merge (create {AST_GCC_ATTRIBUTE_SPECIFIER}.make_from_antlr (a_node, Void))
			end
		end

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_function_declaration (Current)
			if visit_children then
				visit_node (gcc_attribute_specifier, v)
				visit_node (return_type, v)
				visit_node (parameter_list, v)
				visit_list (locals, v)
				visit_node (statements, v)
			end
			v.end_visit_function_declaration (Current)
			v.post_visit (Current)
		end

	ast_twin: like Current
		do
			create Result.make_copy (Current)
		end

	make_copy (a_other: like Current)
		do
			(create {EXCEPTIONS}).raise (generator + ".make_copy")
--			Precursor (a_other)
--			return_type := a_other.return_type.ast_twin
--			return_type.parent := Current
--			parameter_list := a_other.parameter_list.ast_twin
--			parameter_list.parent := Current
--			if a_other.locals /= Void then
--				locals := create {TWO_WAY_LIST [AST_VARIABLE_DECLARATION]}.make
--				across a_other.locals as c loop
--					locals.extend (c.item.ast_twin)
--					locals.last.parent := Current
--				end
--			end
--			if a_other.statements /= Void then
--				statements := a_other.statements.ast_twin
--				statements.parent := Current
--			end
--			has_jump_statements := a_other.has_jump_statements
--			return_statements_removed := a_other.return_statements_removed
--			needs_alloca_factory := a_other.needs_alloca_factory
--			is_longjmp_target := a_other.is_longjmp_target
--			needs_address := a_other.needs_address
		end

invariant
	return_type /= Void
	parameter_list /= Void
	(locals = Void) = (statements = Void)
	oo_preconditions /= Void and then across oo_preconditions as c all c.item.parent = Current end
	oo_postconditions /= Void and then across oo_postconditions as c all c.item.parent = Current end

end
