class
	AST_FLOATING_POINT_CONSTANT

inherit
	AST_EXPRESSION

create
	make,
	make_from_antlr,
	make_copy

feature {NONE}

	make (a_value: STRING; a_type_binding: AST_BASIC_TYPE)
		require
			a_value /= Void
			a_type_binding /= Void -- invariant checks if type is ok
		do
			value := a_value
			type_binding := a_type_binding
		end

	make_from_antlr (node: ANTLR_AST_NODE; init_parent: AST_NODE)
		require
			node /= Void and then node.type = Ast_floating_point_constant
		local
			s: STRING
		do
			parent := init_parent
			s := node.children.first.token_text.as_lower
			if s.ends_with ("f") then
				value := s.substring (1, s.count - 1)
				type_binding := create {AST_BASIC_TYPE}.make (Ast_bt_float)
			elseif s.ends_with ("d") then
				value := s.substring (1, s.count - 1)
				type_binding := create {AST_BASIC_TYPE}.make (Ast_bt_double)
			elseif s.ends_with ("l") then
				value := s.substring (1, s.count - 1)
				type_binding := create {AST_BASIC_TYPE}.make (Ast_bt_long_double)
			else
				value := s
				type_binding := create {AST_BASIC_TYPE}.make (Ast_bt_double)
			end
		end

feature {ANY}

	value: STRING
		-- don't use "NUMERIC"; will reformat the number
		-- e.g. "1.0" to "1" or "1.92" to "1.919999..."

	type_binding: AST_BASIC_TYPE

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_floating_point_constant (Current)
			if visit_children then
				-- no children
			end
			v.end_visit_floating_point_constant (Current)
			v.post_visit (Current)
		end

	ast_twin: like Current
		do
			create Result.make_copy (Current)
		end

	make_copy (other: like Current)
		do
			value := other.value.twin
			type_binding := other.type_binding
		end

invariant
	value /= Void
	type_binding /= Void and then
	(type_binding.type = Ast_bt_float or type_binding.type = Ast_bt_double or type_binding.type = Ast_bt_long_double)

end
