class
	AST_STRING_CONSTANT

inherit
	AST_EXPRESSION

create
	make,
	make_from_antlr,
	make_copy

feature {NONE}

	make (a_value: STRING)
		require
			a_value /= Void
		do
			value := a_value
		ensure
			value = a_value
		end

	make_from_antlr (node: ANTLR_AST_NODE; init_parent: AST_NODE)
		require
			node /= Void and then node.type = Ast_string_constant
		do
			parent := init_parent

			value := node.children.first.token_text
			is_wide := value[1] = 'L'

			-- remove starting and trailing quotes:
			if is_wide then
				value.remove_head (2)
			else
				value.remove_head (1)
			end
			value.remove_tail (1)
		end

feature {ANY}

	value: STRING
		-- The original value from the source. E.g. `A\\\"B\"\n'.

	is_wide: BOOLEAN

	is_eiffel_string: BOOLEAN assign set_is_eiffel_string

	set_is_eiffel_string (a_is_eiffel_string: BOOLEAN)
		do
			is_eiffel_string := a_is_eiffel_string
		end

	type_binding: AST_TYPE
		do
			if is_eiffel_string then
				Result := create {AST_EIFFEL_STRING_TYPE}.make_string
			else
				Result := create {AST_ARRAY_TYPE}.make (create {AST_BASIC_TYPE}.make (Ast_bt_char), Void)
			end
		end

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_string_constant (Current)
			if visit_children then
				-- no children...
			end
			v.end_visit_string_constant (Current)
			v.post_visit (Current)
		end

	ast_twin: like Current
		do
			create Result.make_copy (Current)
		end

	make_copy (other: like Current)
		do
			value := other.value.twin
			is_wide := other.is_wide
			is_eiffel_string := other.is_eiffel_string
		end

invariant
	value /= Void

end
