class
	AST_POINTER_TYPE

inherit
	AST_TYPE
		redefine
			make_copy
		end

create
	make,
	make_from_antlr,
	make_copy

feature {NONE}

	make (a_type: AST_TYPE)
		require
			a_type /= Void and then a_type.parent = Void
		do
			type := a_type
			type.parent := Current
		ensure
			type = a_type
		end

	make_from_antlr (node: ANTLR_AST_NODE; a_type: AST_TYPE; init_parent: AST_NODE)
		require
			node /= Void and then node.type = Ast_pointer_type
			a_type /= Void and then a_type.parent = Void
		local
			pt: AST_POINTER_TYPE
		do
			if not node.children.is_empty and then node.children.last.type = Ast_pointer_type then
				pt := create {AST_POINTER_TYPE}.make (a_type)
				make_from_antlr (node.children.last, pt, init_parent)
			else
				pt := Current
				parent := init_parent
				make (a_type)
			end

			if not node.children.is_empty and then node.children.first.type = Ast_type_qualifier then
				across node.children.first.children as c loop
					inspect
						c.item.type
					when C_kw_const then    pt.is_const := True
					when C_kw_volatile then pt.is_volatile := True
					end
				end
			end
		end

feature {ANY}

	type: AST_TYPE

	non_pointer_type: AST_TYPE
		do
			if attached {AST_POINTER_TYPE} type as tmp then
				Result := tmp.non_pointer_type
			else
				Result := type
			end
		end

	pointer_count: INTEGER
		do
			if attached {AST_POINTER_TYPE} type as tmp then
				Result := tmp.pointer_count + 1
			else
				Result := 1
			end
		end

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_pointer_type (Current)
			if visit_children then
				visit_node (type, v)
			end
			v.end_visit_pointer_type (Current)
			v.post_visit (Current)
		end

	ast_twin: like Current
		do
			create Result.make_copy (Current)
		end

	make_copy (other: like Current)
		do
			Precursor (other)
			type := other.type.ast_twin
			type.parent := Current
		end
feature {ANY} -- AST_TYPE

	is_equal_type (a_type: AST_TYPE): BOOLEAN
		do
			if attached {AST_POINTER_TYPE} a_type as other_type then
				Result := type.is_equal_type (other_type.type)
			end
		end


invariant
	type /= Void

end
