class
	ARITHMETIC_OPERATION_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			end_visit_generic_expression,
			end_visit_unary_expression
		end
	NAME_UTILITIES

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			transformer.tree.accept (Current)
			Result := True
		end

	description : STRING
		do
			Result := "Adapting arithmetic operations where Eiffel doesn't support them directly"
		end

feature {ANY} -- AST_DEFAULT_VISITOR

	end_visit_generic_expression (node: AST_GENERIC_EXPRESSION)
		do
			-- `foo + bar' -> `foo + (int)bar'  where foo is a pointer and bar is a number (but not int)
			if
				(attached {AST_POINTER_TYPE} node.left.type_binding or attached {AST_ARRAY_TYPE} node.left.type_binding) and then
				node.is_additive_expression and then
				attached {AST_BASIC_TYPE} node.right.type_binding as right_type and then
				not (right_type.type = Ast_bt_int or right_type.type = Ast_bt_long)
			then
				replace_node (node.right, create {AST_CAST_EXPRESSION}.make (create {AST_BASIC_TYPE}.make (Ast_bt_int), node.right.ast_twin))

			-- `foo - bar' -> `(foo).distance (bar)'  where foo and bar are pointers
			elseif
				(attached {AST_POINTER_TYPE} node.left.type_binding or attached {AST_ARRAY_TYPE} node.left.type_binding) and then
				node.type = C_op_minus and then
				(attached {AST_POINTER_TYPE} node.right.type_binding or attached {AST_ARRAY_TYPE} node.right.type_binding)
			then
				replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (
						<<
							create {AST_PARENTHESIZED_EXPRESSION}.make (node.left.ast_twin),
							".distance (",
							create {AST_PARENTHESIZED_EXPRESSION}.make (node.right.ast_twin),
							")"
						>>,
						node.type_binding
					))

			-- `foo << bar' -> `foo << (int)bar' where bar is not of type int
			elseif
				node.is_shift_expression and then
				attached {AST_BASIC_TYPE} node.right.type_binding as right_type and then
				not (right_type.type = Ast_bt_int or right_type.type = Ast_bt_long)
			then
				replace_node (node.right, create {AST_CAST_EXPRESSION}.make (
						create {AST_BASIC_TYPE}.make (Ast_bt_int),
						create {AST_PARENTHESIZED_EXPRESSION}.make (node.right.ast_twin)
					))
			end
		end

	end_visit_unary_expression (node: AST_UNARY_EXPRESSION)
		local
			int_type: INTEGER_32
		do
			-- `-ui' -> `(unsigned int)-(int)ui' where ui is an unsigned integer
			if
				node.type = C_op_minus and then
				attached {AST_BASIC_TYPE} node.expression.type_binding as bt and then
				(
					bt.type = Ast_bt_unsigned_char or bt.type = Ast_bt_unsigned_short or
					bt.type = Ast_bt_unsigned_int or bt.type = Ast_bt_unsigned_long or
					bt.type = Ast_bt_unsigned_long_long
				)
			then
				inspect
					bt.type
				when Ast_bt_unsigned_char then      int_type := Ast_bt_char
				when Ast_bt_unsigned_short then     int_type := Ast_bt_short
				when Ast_bt_unsigned_int then       int_type := Ast_bt_int
				when Ast_bt_unsigned_long then      int_type := Ast_bt_long
				when Ast_bt_unsigned_long_long then int_type := Ast_bt_long_long
				end
				-- `-ui' -> `-(int)ui'
				replace_node (node.expression, create {AST_CAST_EXPRESSION}.make (
						create {AST_BASIC_TYPE}.make (int_type),
						node.expression.ast_twin
					))
				-- `-(int)ui' -> `(unsigned int)-(int)ui'
				replace_node (node, create {AST_CAST_EXPRESSION}.make (bt.ast_twin, node.ast_twin))
			end
		end

end
