class
	EIFFEL_FILE_WRITER

inherit
	TRANSFORMATION_STEP
		redefine
			emit_source
		end
	NAME_UTILITIES
	ARGUMENTS

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		local
			l_main_function: AST_FUNCTION_DECLARATION
			l_system_dir: DIRECTORY
			l_assembler_statements: LIST [AST_ASSEMBLER_STATEMENT]
			l_base_class_name, l_launcher_class_name, l_assembler_class_name: STRING_8
		do
			across
				transformer.tree.declarations as c
			until
				l_main_function /= Void
			loop
				if
					attached {AST_FUNCTION_DECLARATION} c.item as tmp and then
					tmp.identifier.is_equal ("main")
				then
					l_main_function := tmp
				end
			end
			if l_main_function /= Void then
				log ("Function %"main%" found, compiling as program.")
			else
				log ("Function %"main%" not found, compiling as library.")
			end

			across transformer.tree.declarations as c loop
				if attached {AST_STRUCT_DECLARATION} c.item as sd then
					if sd.source_info.is_project_source then
						write_class (transformer.project_settings.output_dir, sd)
					else
						if l_system_dir = Void then
							create l_system_dir.make (transformer.project_settings.output_dir.name + "/system")
							if not l_system_dir.exists then
								l_system_dir.create_dir
							end
						end
						write_class (l_system_dir, sd)
					end
				elseif attached {AST_ENUM_DECLARATION} c.item as ed then
					if ed.source_info.is_project_source then
						write_constant_class (transformer.project_settings.output_dir, ed)
					else
						if l_system_dir = Void then
							create l_system_dir.make (transformer.project_settings.output_dir.name + "/system")
							if not l_system_dir.exists then
								l_system_dir.create_dir
							end
						end
						write_constant_class (l_system_dir, ed)
					end
				end
			end

			l_assembler_statements := (create {ASSEMBLER_TRANSFORMER}).assembler_statements

			l_base_class_name := eiffel_class_name (transformer.tree.source_file_name)
			l_launcher_class_name := l_base_class_name + "_LAUNCHER"
			l_assembler_class_name := l_base_class_name + "_ASSEMBLER"

			-- adding project framework classes
			write_ecf_file (l_launcher_class_name, l_base_class_name, l_main_function /= Void)
			if l_main_function /= Void then
				write_project_launcher_class (l_main_function, l_launcher_class_name)
			end
			if not l_assembler_statements.is_empty then
				write_assembler_class (l_assembler_statements, l_assembler_class_name)
			end
			write_header_file (l_base_class_name)
			if attached resources_directory as rd then
				copy_utilities (rd.name + "/ce_utilities", "ce_utilities")
				copy_utilities (rd.name + "/ce_utilities/impl", "ce_utilities/impl")
				if {PLATFORM}.is_unix then
					copy_utilities (rd.name + "/include/linux", "include")
					copy_utilities (rd.name + "/objects/linux", "objects")
				elseif {PLATFORM}.is_windows then
					copy_utilities (rd.name + "/include/windows", "include")
					copy_utilities (rd.name + "/objects/windows", "objects")
				end
				write_ce_settings_class (create {DIRECTORY}.make (transformer.project_settings.output_dir.name + "/ce_utilities"))
			else
				log ("Can't copy utilities, %"resources%" directory not found.")
			end

			Result := True
		end

	description: STRING
		do
			Result := "Writing Eiffel File(s)"
		end

	emit_source (text: EV_TEXT)
		do
			text.set_text ("All source files are in %"" + transformer.project_settings.output_dir.name + "%".")
		end

feature {NONE}

	write_class (a_dir: DIRECTORY; a_class: AST_STRUCT_DECLARATION)
		require
			a_dir /= Void and then a_dir.exists
			a_class /= Void
		local
			l_file: PLAIN_TEXT_FILE
			l_oo_struct_writer: STRUCT_WRITER_OO
			l_c_struct_writer: STRUCT_WRITER_C
			l_buffer, l_static_buffer: STRING_8
		do
			if not transformer.project_settings.disable_oo_structs then
				create l_oo_struct_writer.make (a_class)
				l_oo_struct_writer.write
				l_buffer := l_oo_struct_writer.buffer
				l_static_buffer := l_oo_struct_writer.static_buffer
			else
				create l_c_struct_writer.make (a_class)
				l_c_struct_writer.write
				l_buffer := l_c_struct_writer.buffer
				l_static_buffer := l_c_struct_writer.static_buffer
			end

			if l_buffer /= Void then
				create l_file.make_open_write (a_dir.name + "/" + eiffel_file_name (a_class.identifier))
				l_file.put_string (l_buffer)
				l_file.flush
				l_file.close
			end
			if l_static_buffer /= Void then
				create l_file.make_open_write (a_dir.name + "/" + eiffel_file_name (a_class.identifier + "_static"))
				l_file.put_string (l_static_buffer)
				l_file.flush
				l_file.close
			end
		end

	write_constant_class (a_dir: DIRECTORY; a_class: AST_ENUM_DECLARATION)
		require
			a_dir /= Void and then a_dir.exists
			a_class /= Void
		local
			l_file: PLAIN_TEXT_FILE
		do
			create l_file.make_open_write (a_dir.name + "/" + eiffel_file_name (a_class.identifier))
			l_file.put_string (a_class.eiffel_code)
			l_file.flush
			l_file.close
		end

	write_assembler_class (assembler_statements: LIST [AST_ASSEMBLER_STATEMENT]; assembler_class_name: STRING)
		require
			assembler_statements /= Void and then not assembler_statements.is_empty
		local
			file: PLAIN_TEXT_FILE
			c_creator: AST_C_CREATOR
			l_type: AST_TYPE
		do
			create file.make_open_write (transformer.project_settings.output_dir.name + "/" + eiffel_file_name (assembler_class_name))
			file.put_string ("class%N")
			file.put_string ("	" + assembler_class_name + "%N%N")

			file.put_string ("feature {NONE}%N%N")

			create c_creator.make
			c_creator.indent := 4
			across assembler_statements as c loop
				file.put_string ("	asm" + c.target_index.out)
				if not c.item.accessed_variables.is_empty then
					file.put_string (" (")
					across c.item.accessed_variables as c2 loop
						if c2.target_index > 1 then
							file.put_string ("; ")
						end
						file.put_string ("op")
						file.put_integer_32 (c2.target_index)
						file.put_string (": ")
						file.put_string (c2.item.type_binding.eiffel_code)
					end
					file.put_string (")%N")
					file.put_string ("		do%N")
					file.put_string ("			c_asm" + c.target_index.out + " (")
					across c.item.accessed_variables as c2 loop
						if c2.target_index > 1 then
							file.put_string (", ")
						end
						file.put_string ("op")
						file.put_integer_32 (c2.target_index)
						file.put_string (".c_pointer")
					end
					file.put_string (")%N")
					file.put_string ("		end%N%N")

					file.put_string ("	c_asm" + c.target_index.out)
					file.put_string (" (")
					across c.item.accessed_variables as c2 loop
						if c2.target_index > 1 then
							file.put_string ("; ")
						end
						file.put_string ("op")
						file.put_integer_32 (c2.target_index)
						file.put_string (": POINTER")
					end
					file.put_string (")")
				end
				file.put_string ("%N")
				file.put_string ("		external%N")
				file.put_string ("			%"C inline use <" + c_header_file (transformer.tree) + ">%"%N")
				file.put_string ("		alias%N")
				file.put_string ("			%"[%N")

				if not c.item.accessed_variables.is_empty then
					across c.item.accessed_variables as c2 loop
						if attached {AST_VARIABLE_DECLARATION} c2.item.declaration_binding as vd then
							l_type := ({AST_ARRAY_TYPE} #? vd.type).type

							-- int x = *(int *)$opX;
							file.put_string ("				")
							if attached {AST_ARRAY_TYPE} l_type as at then
								file.put_string ((create {AST_VARIABLE_DECLARATION}.make (at.as_pointer, vd.orig_c_identifier, Void)).c_code)
							else
								file.put_string ((create {AST_VARIABLE_DECLARATION}.make (l_type.ast_twin, vd.orig_c_identifier, Void)).c_code)
							end

							file.put_string (" = ")

							if attached {AST_ARRAY_TYPE} l_type then
								-- we used `arr.address' what is not really an address; don't dereference it
							else
								file.put_string ("*")
							end
							file.put_string ("(")
							if attached {AST_ARRAY_TYPE} l_type as at then
								file.put_string (at.as_pointer.c_code)
								-- no additional pointer needed; see comment above
							else
								file.put_string (l_type.c_code)
								file.put_string (" *")
							end
							file.put_string (")$op")
							file.put_integer_32 (c2.target_index)
							file.put_string (";%N")
						else
							file.put_string ("?" + c2.item.declaration_binding.generator + "?")
						end
					end
					file.put_string ("%N")
				end

				c.item.accept (c_creator)
				file.put_string ("				")
				file.put_string (c_creator.buffer)
				file.put_string ("%N")
				c_creator.buffer.wipe_out

				if not c.item.accessed_variables.is_empty then
					file.put_string ("%N")
					across c.item.accessed_variables as c2 loop
						if attached {AST_VARIABLE_DECLARATION} c2.item.declaration_binding as vd then
							l_type := ({AST_ARRAY_TYPE} #? vd.type).type

							if attached {AST_ARRAY_TYPE} l_type then
								-- no need to save back; see comment above
							else
								-- *(int *)$x = x;
								file.put_string ("				*(")
								file.put_string (remove_const_qualifiers (l_type).c_code)
								file.put_string (" *)$op")
								file.put_integer_32 (c2.target_index)
								file.put_string (" = ")
								file.put_string (vd.orig_c_identifier)
								file.put_string (";%N")
							end
						else
							file.put_string ("?" + c2.item.declaration_binding.generator + "?")
						end
					end
				end

				file.put_string ("			]%"%N")
				file.put_string ("		end%N%N")
			end

			file.put_string ("end%N")
			file.flush
			file.close
		end

	write_ecf_file (launcher_class_name: STRING; base_class_name: STRING; is_project: BOOLEAN)
		require
			launcher_class_name /= Void
			base_class_name /= Void
		local
			file: PLAIN_TEXT_FILE
		do
			create file.make_open_write (transformer.project_settings.output_dir.name + "/" + base_class_name.as_lower + ".ecf")
			file.put_string ("<?xml version=%"1.0%" encoding=%"ISO-8859-1%"?>%N")
			file.put_string ("<system name=%"" + base_class_name.as_lower + "%" ")
			if not is_project then
				file.put_string ("library_target=%"" + base_class_name.as_lower + "%" ")
			end
			file.put_string ("uuid=%"" + (create {UUID_GENERATOR}).generate_uuid.out + "%">%N")
			file.put_string ("	<target name=%"" + base_class_name.as_lower + "%">%N")
			if is_project then
				file.put_string ("		<root class=%"" + launcher_class_name + "%" feature=%"make%"/>%N")
			else
				file.put_string ("		<root all_classes=%"true%"/>%N")
			end
			file.put_string ("		<option warning=%"true%" syntax=%"provisional%">%N")
			file.put_string ("			<assertions precondition=%"true%" postcondition=%"true%" check=%"true%" invariant=%"true%" loop=%"true%" supplier_precondition=%"true%"/>%N")
			file.put_string ("		</option>%N")
			file.put_string ("		<setting name=%"console_application%" value=%"true%"/>%N")
			file.put_string ("		<external_include location=%"&quot;$ECF_CONFIG_PATH/include&quot;%"/>%N")
			across transformer.project_settings.include_dirs as c loop
				file.put_string ("		<external_include location=%"&quot;" + c.item + "&quot;%"/>%N")
			end
			if {PLATFORM}.is_unix then
				file.put_string ("		<external_object location=%"&quot;$ECF_CONFIG_PATH/objects/libffi.a&quot;%"/>%N")
			elseif {PLATFORM}.is_windows then
				file.put_string ("		<external_object location=%"&quot;$ECF_CONFIG_PATH/objects/libffi.lib&quot;%"/>%N")
			end
			across transformer.project_settings.libraries as c loop
				file.put_string ("		<external_library location=%"&quot;" + c.item + "&quot;%"/>%N")
			end
			if is_project then
				file.put_string ("		<precompile name=%"base_pre%" location=%"$ISE_PRECOMP\base.ecf%"/>%N")
			end
			file.put_string ("		<library name=%"base%" location=%"$ISE_LIBRARY\library\base\base.ecf%"/>%N")
			file.put_string ("		<cluster name=%"" + base_class_name.as_lower + "%" location=%".\%" recursive=%"true%">%N")
			file.put_string ("			<file_rule>%N")
			file.put_string ("				<exclude>/EIFGENs$</exclude>%N")
			file.put_string ("				<exclude>/CVS$</exclude>%N")
			file.put_string ("				<exclude>/.svn$</exclude>%N")
			file.put_string ("			</file_rule>%N")
			file.put_string ("		</cluster>%N")
			file.put_string ("	</target>%N")
			file.put_string ("</system>%N")
			file.flush
			file.close
		end

	write_project_launcher_class (main_function: AST_FUNCTION_DECLARATION; launcher_class_name: STRING_8)
		require
			main_function /= Void and then main_function.identifier.is_equal ("main")
			launcher_class_name /= Void
		local
			l_has_return, l_has_argc, l_has_argc_argv: BOOLEAN
			l_file: PLAIN_TEXT_FILE
		do
			l_has_return := not (attached {AST_BASIC_TYPE} main_function.return_type as bt and then bt.type = Ast_bt_void)
			inspect main_function.parameter_list.parameters.count
			when 0 then
				-- no args
			when 1 then
				l_has_argc := True
			when 2 then
				l_has_argc_argv := True
			end

			create l_file.make_open_write (transformer.project_settings.output_dir.name + "/" + eiffel_file_name (launcher_class_name))
			l_file.put_string ("class%N")
			l_file.put_string ("	" + launcher_class_name + "%N%N")

			if l_has_argc or l_has_argc_argv then
				l_file.put_string ("inherit%N")
				l_file.put_string ("	ARGUMENTS%N")
				if l_has_argc_argv then
					l_file.put_string ("	CE_UTILITIES%N")
				end
				l_file.put_string ("%N")
			end

			l_file.put_string ("create%N")
			l_file.put_string ("	make%N%N")

			l_file.put_string ("feature {NONE} -- Initialization%N%N")

			l_file.put_string ("	make%N")
			if l_has_return or l_has_argc_argv then
				l_file.put_string ("		local%N")
				if l_has_argc_argv then
					l_file.put_string ("			argv: CE_ARRAY [CE_POINTER [INTEGER_8]]%N")
				end
				if l_has_return then
					l_file.put_string ("			return_value: INTEGER_32%N")
				end
			end
			l_file.put_string ("		do%N")
			if l_has_argc_argv then
				l_file.put_string ("			create {CE_POINTER_ARRAY [INTEGER_8]} argv.make (<<argument_count + 2>>)%N")
				l_file.put_string ("			across argument_array as c loop%N")
				l_file.put_string ("				argv[c.target_index] := ce_string (c.item)%N")
				l_file.put_string ("			end%N")
			end

			l_file.put_string ("			")
			if l_has_return then
				l_file.put_string ("return_value := ")
			end
			l_file.put_string ("(create {" + main_function.oo_class.identifier + "_STATIC})" + ".main")
			if l_has_argc or l_has_argc_argv then
				l_file.put_string (" (argument_count + 1")
				if l_has_argc_argv then
					l_file.put_string (", argv")
				end
				l_file.put_string (")")
			end
			l_file.put_string ("%N")

			if l_has_return then
				l_file.put_string ("			if return_value /= 0 then%N")
				l_file.put_string ("				(create {EXCEPTIONS}).die (return_value)%N")
				l_file.put_string ("			end%N")
			end

			l_file.put_string ("		end%N%N")

			l_file.put_string ("end%N")
			l_file.flush
			l_file.close
		end

	write_header_file (base_class_name: STRING)
		require
			base_class_name /= Void
		local
			d: DIRECTORY
			f: PLAIN_TEXT_FILE
			set: HASH_SET [STRING]
			printed: BOOLEAN
		do
			create d.make (transformer.project_settings.output_dir.name + "/include")
			if not d.exists then
				d.create_dir
			end
			create f.make_open_write (d.name + "/" + c_header_file (transformer.tree))
			f.put_string ("#ifndef " + base_class_name + "_H%N")
			f.put_string ("#define " + base_class_name + "_H%N%N")

			-- header files
			create set.make (16)
			across transformer.tree.declarations as c loop
				if
					not c.item.source_info.is_pure_oo and then
					c.item.source_info.is_system_source and then
					not c.item.source_info.source_file.is_equal ("gcc_builtin.h")
				then
					set.extend (c.item.source_info.source_file)
				end
			end
			f.put_string ("#include <ffi.h>%N")
			across set as c loop
				f.put_string ("#include <" + c.item + ">%N")
			end
			f.put_string ("%N")

			-- Some header files might have macros that screw with our C code (happens in "links").
			-- "undef"ing stuff here:
			printed := False
			create set.make (1000)
			across transformer.tree.declarations as c loop
				if attached {AST_STRUCT_DECLARATION} c.item as tmp then
					across tmp.fields as c2 loop
						if attached c2.item.identifier as identifier then
							set.extend (identifier)
						end
					end
				elseif attached {AST_FUNCTION_DECLARATION} c.item as tmp and then tmp.is_extern then
					set.extend (tmp.identifier)
					across tmp.parameter_list.parameters as c2 loop
						if attached {AST_VARIABLE_DECLARATION} c2.item as tmp2 then
							set.extend (tmp2.orig_c_identifier)
						end
					end
				end
			end
			across set as c loop
				f.put_string ("#undef ")
				f.put_string (c.item)
				f.put_string ("%N")
				printed := True
			end

			-- enums
			if printed then f.put_string ("%N") end
			printed := False
			across transformer.tree.declarations as c loop
				if
					attached {AST_ENUM_DECLARATION} c.item as tmp and then
					tmp.source_info.is_project_source
				then
					f.put_string (tmp.c_code)
					f.put_string ("%N")
					printed := True
				end
			end

			-- struct forward declarations
			if printed then f.put_string ("%N") end
			printed := False
			across transformer.tree.declarations as c loop
				if
					attached {AST_STRUCT_DECLARATION} c.item as tmp and then
					not tmp.source_info.is_pure_oo and then
					(tmp.source_info.is_project_source or tmp.orig_c_identifier.starts_with ("__anon"))
				then
					if tmp.is_union then
						f.put_string ("union ")
					else
						f.put_string ("struct ")
					end
					f.put_string (tmp.orig_c_identifier)
					f.put_string (";%N")
					printed := True
				end
			end

			-- typedefs
			if printed then f.put_string ("%N") end
			printed := False
			across transformer.tree.declarations as c loop
				if
					attached {AST_STRUCT_DECLARATION} c.item as tmp and then
					not tmp.source_info.is_pure_oo and then
					tmp.source_info.is_project_source and then
					tmp.orig_c_identifier.starts_with ("__anon") and then
					tmp.typedef_name /= Void
				then
					f.put_string ("typedef ")
					if tmp.is_union then
						f.put_string ("union ")
					else
						f.put_string ("struct ")
					end
					f.put_string (tmp.orig_c_identifier + " " + tmp.typedef_name + ";%N")
					printed := True
				end
			end

			-- structs
			if printed then f.put_string ("%N") end
			printed := False
			across transformer.tree.declarations as c loop
				if
					attached {AST_STRUCT_DECLARATION} c.item as tmp and then
					not tmp.source_info.is_pure_oo and then
					(tmp.source_info.is_project_source or tmp.orig_c_identifier.starts_with ("__anon"))
				then
					f.put_string (tmp.c_code)
					f.put_string ("%N")
					printed := True
				end
			end

			-- external features from project headers
			if printed then f.put_string ("%N") end
			printed := False
			across transformer.tree.declarations as c loop
				if
					attached {AST_EIFFEL_FEATURE} c.item as tmp and then
					tmp.is_extern and then
					(
						tmp.source_info.is_project_source or
						tmp.orig_c_identifier.is_equal ("__builtin_va_copy")
					)
				then
					f.put_string (tmp.c_code)
					if attached {AST_VARIABLE_DECLARATION} tmp then
						f.put_string (";")
					end
					f.put_string ("%N")
					printed := True
				end
			end

			if printed then f.put_string ("%N") end
			f.put_string ("#endif%N")
			f.flush
			f.close
		end

	resources_directory: DIRECTORY
		local
			path: STRING
		do
			create Result.make ("../resources") -- we're in EiffelStudio
			if not Result.exists then
				create Result.make ("resources") -- it's a release
				if not Result.exists then
					path := Command_name -- e.g. "/foo/bar/C2Eif", "bar/C2Eif", "C2Eif"
					path := path.substring (1, path.last_index_of ('/', path.count)) -- "/foo/bar/", "bar/", ""
					create Result.make (path + "resources")
					if not Result.exists then
						Result := Void
					end
				end
			end
		end

	copy_utilities (src_path, dest_path: STRING)
		require
			src_path /= Void
			dest_path /= Void
		local
			src_directory, dest_directory: DIRECTORY
			src_file, dest_file: RAW_FILE
		do
			create src_directory.make (src_path)
			if src_directory.exists then
				create dest_directory.make (transformer.project_settings.output_dir.name + "/" + dest_path)
				if not dest_directory.exists then
					dest_directory.create_dir
				end

				across src_directory.linear_representation as c loop
					create src_file.make (src_directory.name + "/" + c.item)
					if not src_file.is_directory then -- ignore `.' and `..'
						src_file.open_read
						create dest_file.make_open_write (dest_directory.name + "/" + c.item)
						src_file.copy_to (dest_file)
						dest_file.flush
						dest_file.close
						src_file.close
					end
				end
			else
				log ("Can't copy utilities from %"" + src_path + "%": directory doesn't exist")
			end
		end

	write_ce_settings_class (dest_path: DIRECTORY)
		require
			dest_path /= Void and then dest_path.exists
		local
			f: PLAIN_TEXT_FILE
		do
			create f.make_open_write (dest_path.name + "/ce_settings.e")
			f.put_string ("class%N")
			f.put_string ("	CE_SETTINGS%N")
			f.put_string ("%N")
			f.put_string ("feature {ANY}%N")
			f.put_string ("%N")
			f.put_string ("	Disable_bounds_checking: BOOLEAN = " + transformer.project_settings.disable_bounds_checking.out + "%N")
			f.put_string ("%N")
			f.put_string ("end%N")
			f.flush
			f.close
		end

	remove_const_qualifiers (a_node: AST_NODE): AST_NODE
		local
			l_visitor: AST_AGENT_VISITOR
		do
			Result := a_node.ast_twin

			create l_visitor
			l_visitor.agents.post_visit := agent (ag_node: AST_NODE)
				do
					if attached {AST_TYPE} ag_node as tmp and then tmp.is_const then
						tmp.is_const := False
					end
				end
			Result.accept (l_visitor)
		end

end
