class
	FORWARD_DECLARATIONS_REMOVER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			end_visit_struct_type
		end

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		local
			functions: HASH_TABLE [TWO_WAY_LIST [AST_DECLARATION], STRING]
			variables: HASH_TABLE [TWO_WAY_LIST [AST_DECLARATION], STRING]
			enums: HASH_TABLE [TWO_WAY_LIST [AST_DECLARATION], STRING]
			structs: HASH_TABLE [TWO_WAY_LIST [AST_DECLARATION], STRING]
			cur_map: HASH_TABLE [TWO_WAY_LIST [AST_DECLARATION], STRING]
		do
			create functions.make (16)
			create variables.make (16)
			create enums.make (16)
			create structs.make (16)

			-- Save all actual and forward declarations per declaration into lists:
			across transformer.tree.declarations as c loop
				if attached {AST_FUNCTION_DECLARATION} c.item as tmp then
					cur_map := functions
				elseif attached {AST_VARIABLE_DECLARATION} c.item as tmp then
					cur_map := variables
				elseif attached {AST_ENUM_DECLARATION} c.item as tmp then
					cur_map := enums
				elseif attached {AST_STRUCT_DECLARATION} c.item as tmp then
					cur_map := structs
				else
					log ("Unknown global: " + c.item.generator)
					cur_map := Void
				end
				if cur_map /= Void then
					if not cur_map.has (c.item.identifier) then
						cur_map.extend (create {TWO_WAY_LIST [AST_DECLARATION]}.make, c.item.identifier)
					end
					cur_map[c.item.identifier].extend (c.item)
				end
			end

			-- Delete the forward declarations (not straight forward for variables and structs):
			across functions as c loop
				delete_forwards (c.item)
			end
			across variables as c loop
				delete_forwards (c.item)
			end
			across enums as c loop
				delete_forwards (c.item)
			end

			-- struct can exists without declaration. E.g. `struct _XDisplay;'.
			-- It can then be used as a "typed" pointer. E.g. `struct _XDisplay *display;'
			-- Below we delete structs without declaration and adapt the usages to void pointers. E.g. `void *display;'
			create struct_forwards_only.make (16)
			across structs as c loop
				delete_forwards (c.item)
			end
			if not struct_forwards_only.is_empty then
				transformer.tree.accept (Current)
			end
			struct_forwards_only := Void

			Result := True
		end

	description: STRING
		do
			Result := "Removing forward declarations"
		end

feature {NONE} -- AST_DEFAULT_VISITOR

	end_visit_struct_type (node: AST_STRUCT_TYPE)
		do
			if struct_forwards_only.has (node.identifier) then
				replace_node (node, create {AST_BASIC_TYPE}.make (Ast_bt_void))
			end
		end

feature {NONE}

	struct_forwards_only: HASH_SET [STRING]

	delete_forwards (declarations: LIST [AST_DECLARATION])
		local
			has_declaration: BOOLEAN
		do
			across declarations as c loop
				if is_forward_declaration (c.item) then
					if not has_declaration and c.target_index = declarations.count then
						handle_forward_declaration_only (c.item)
					else
						delete_node (c.item)
					end
				else
					has_declaration := True
				end
			end
		end

	is_forward_declaration (node: AST_DECLARATION): BOOLEAN
		do
			if attached {AST_FUNCTION_DECLARATION} node as tmp then
				Result := tmp.is_forward_declaration
			elseif attached {AST_VARIABLE_DECLARATION} node as tmp then
				Result := (tmp.initializer = Void)
			elseif attached {AST_ENUM_DECLARATION} node as tmp then
				Result := tmp.is_forward_declaration
			elseif attached {AST_STRUCT_DECLARATION} node as tmp then
				Result := tmp.is_forward_declaration
			else
				log ("Unknown global: " + node.generator)
			end
		end

	handle_forward_declaration_only (last_forward_declaration: AST_DECLARATION)
		do
			if attached {AST_FUNCTION_DECLARATION} last_forward_declaration then
				log ("forward function declaration only: " + last_forward_declaration.identifier)
			elseif attached {AST_VARIABLE_DECLARATION} last_forward_declaration then
				-- that's ok...
			elseif attached {AST_ENUM_DECLARATION} last_forward_declaration then
				log ("forward enum declaration only: " + last_forward_declaration.identifier)
			elseif attached {AST_STRUCT_DECLARATION} last_forward_declaration then
				delete_node (last_forward_declaration)
				struct_forwards_only.extend (last_forward_declaration.identifier)
			else
				log ("Unknown global: " + last_forward_declaration.generator)
			end
		end

end
