class
	INCREMENT_DECREMENT_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			visit_postfix_expression,
			visit_unary_expression
		end

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			transformer.tree.accept (Current)
			Result := True
		end

	description : STRING
		do
			Result := "Transforming increment (++) and decrement (--)"
		end

feature {ANY} -- AST_DEFAULT_VISITOR

	visit_postfix_expression (node: AST_POSTFIX_EXPRESSION): BOOLEAN
		do
			if attached {AST_EXPRESSION_STATEMENT} node.parent then
				if node.type = C_op_increment then
					create_assignment_expression (node.main_expression, C_op_plus)
				elseif node.type = C_op_decrement then
					create_assignment_expression (node.main_expression, C_op_minus)
				end
			end
		end

	visit_unary_expression (node: AST_UNARY_EXPRESSION): BOOLEAN
		do
			if attached {AST_EXPRESSION_STATEMENT} node.parent then
				if node.type = C_op_increment then
					create_assignment_expression (node.expression, C_op_plus)
				elseif node.type = C_op_decrement then
					create_assignment_expression (node.expression, C_op_minus)
				end
			end
		end

feature {NONE}

	create_assignment_expression (expression: AST_EXPRESSION; type: INTEGER)
		require
			expression /= Void
			type = C_op_plus or type = C_op_minus
		do
			replace_node (expression.parent, create {AST_ASSIGNMENT_EXPRESSION}.make (
					expression.non_parentheses_expression.ast_twin,
					create {AST_GENERIC_EXPRESSION}.make (
							type,
							expression.ast_twin,
							create {AST_DECIMAL_CONSTANT}.make (1, create {AST_BASIC_TYPE}.make (Ast_bt_int)))))
		end

end
