class
	OO_FINISH

inherit
	TRANSFORMATION_STEP

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			add_system_superclasses
			if not transformer.project_settings.disable_oo_reengineering then
				resolve_name_clashes
				if not transformer.project_settings.disable_oo_structs then
					update_class_fields_where_necessary
				end
			end
			Result := True
		end

	description: STRING
		do
			Result := "Finishing up O-O reengineering"
		end

feature {NONE}

	add_system_superclasses
		local
			l_cur_data: TUPLE [struct: AST_STRUCT_DECLARATION; is_static: BOOLEAN]
			l_visitor: AST_AGENT_VISITOR
		do
			create l_cur_data
			create l_visitor
			l_visitor.agents.end_visit_function_call := agent (ag_node: AST_FUNCTION_CALL; ag_cur_data: TUPLE [struct: AST_STRUCT_DECLARATION; is_static: BOOLEAN])
				do
					if
						attached ag_node.declaration_binding as called_function and then
						called_function.oo_class.source_info.is_system_source
					then
						if ag_cur_data.is_static then
							ag_cur_data.struct.oo_static_superclasses.extend (called_function.oo_class)
						elseif not ag_cur_data.struct.oo_static_superclasses.has (called_function.oo_class) then
							ag_cur_data.struct.oo_superclasses.extend (called_function.oo_class)
						end
					end
				end (?, l_cur_data)
			l_visitor.agents.end_visit_variable_expression := agent (ag_node: AST_VARIABLE_EXPRESSION; ag_cur_data: TUPLE [struct: AST_STRUCT_DECLARATION; is_static: BOOLEAN])
				do
					if
						attached ag_node.declaration_binding as called_variable and then
						attached {AST_TRANSLATION_UNIT} called_variable.parent and then
						called_variable.oo_class.source_info.is_system_source
					then
						if ag_cur_data.is_static then
							ag_cur_data.struct.oo_static_superclasses.extend (called_variable.oo_class)
						elseif not ag_cur_data.struct.oo_static_superclasses.has (called_variable.oo_class) then
							ag_cur_data.struct.oo_superclasses.extend (called_variable.oo_class)
						end
					end
				end (?, l_cur_data)

			across transformer.tree.declarations as c loop
				if
					attached {AST_STRUCT_DECLARATION} c.item as sd and then
					sd.source_info.is_project_source
				then
					l_cur_data.struct := sd
					l_cur_data.is_static := True
					across sd.oo_static_features as c2 loop
						c2.item.accept (l_visitor)
					end
					l_cur_data.is_static := False
					across sd.oo_features as c2 loop
						c2.item.accept (l_visitor)
					end
				end
			end
		end

feature {NONE}

	resolve_name_clashes
		local
			l_existing_identifiers: HASH_SET [STRING_8]
		do
			across transformer.tree.declarations as declaration_cursor loop
				if attached {AST_STRUCT_DECLARATION} declaration_cursor.item as sd then
					create l_existing_identifiers.make (16)
					across sd.oo_superclasses as struct_cursor loop
						across struct_cursor.item.fields as field_cursor loop
							if attached field_cursor.item.identifier as identifier then
								l_existing_identifiers.extend (identifier.as_lower)
							end
						end
						across struct_cursor.item.oo_features as feature_cursor loop
							l_existing_identifiers.extend (feature_cursor.item.identifier.as_lower)
						end
						across struct_cursor.item.oo_static_features as feature_cursor loop
							l_existing_identifiers.extend (feature_cursor.item.identifier.as_lower)
						end
					end
					across sd.oo_static_superclasses as struct_cursor loop
						across struct_cursor.item.fields as field_cursor loop
							if attached field_cursor.item.identifier as identifier then
								l_existing_identifiers.extend (identifier.as_lower)
							end
						end
						across struct_cursor.item.oo_features as feature_cursor loop
							l_existing_identifiers.extend (feature_cursor.item.identifier.as_lower)
						end
						across struct_cursor.item.oo_static_features as feature_cursor loop
							l_existing_identifiers.extend (feature_cursor.item.identifier.as_lower)
						end
					end

					across sd.oo_defined_fields as field_cursor loop
						if attached field_cursor.item.identifier as identifier then
							check_identifier (identifier, l_existing_identifiers)
						end
					end
					across sd.oo_features as feature_cursor loop
						check_identifier (feature_cursor.item.identifier, l_existing_identifiers)
					end
					across sd.oo_static_features as feature_cursor loop
						check_identifier (feature_cursor.item.identifier, l_existing_identifiers)
					end
				end
			end
		end

	check_identifier (a_identifier: STRING_8; a_existing_identifiers: HASH_SET [STRING_8])
		require
			a_identifier /= Void
			a_existing_identifiers /= Void
		local
			l_identifier: STRING_8
			l_i: INTEGER_32
		do
			from
				l_identifier := a_identifier.as_lower
				l_i := 1
			until
				not a_existing_identifiers.has (l_identifier)
			loop
				l_i := l_i + 1
				l_identifier := a_identifier.as_lower + l_i.out
			end
			if l_i > 1 then
				a_identifier.append_integer (l_i)
			end
			a_existing_identifiers.extend (l_identifier.as_lower)
		end

feature {NONE}

	update_class_fields_where_necessary
			-- Functions that changed a struct pointer argument to Current might become
			-- inconsistent after function calls or assignments if Current is aliased
			-- somewhere else (see oo_reeng_struct_pointer_arg_aliasing.c).
			-- We therefore add calls to `update_class_fields' after assignments and function calls.
		local
			l_update_statements: HASH_SET [AST_STATEMENT]
			l_visitor: AST_AGENT_VISITOR
		do
			-- collect all the statements to avoid making duplicate calls to `update_class_fields' (e.g. from `tmp = foo()')
			create l_update_statements.make (16)
			create l_visitor
			l_visitor.agents.end_visit_function_call := agent (ag_node: AST_FUNCTION_CALL; ag_update_statements: HASH_SET [AST_STATEMENT])
				do
					-- TODO; only add call if necessary
					ag_update_statements.extend ({AST_STATEMENT} #? ag_node.parent_of_type ({AST_STATEMENT}))
				end (?, l_update_statements)
			l_visitor.agents.end_visit_assignment_expression := agent (ag_node: AST_ASSIGNMENT_EXPRESSION; ag_update_statements: HASH_SET [AST_STATEMENT])
				do
					-- TODO; can more calls be removed?
					if
						not attached {AST_VARIABLE_EXPRESSION} ag_node.target and
						not (attached {AST_EIFFEL_EXPRESSION} ag_node.target as ee and then ee.mixed.count = 1 and then ee.is_string (1, "Result"))
					then
						ag_update_statements.extend ({AST_STATEMENT} #? ag_node.parent_of_type ({AST_STATEMENT}))
					end
				end (?, l_update_statements)

			across transformer.tree.declarations as c loop
				if
					attached {AST_FUNCTION_DECLARATION} c.item as fd and then
					fd.source_info.is_project_source and then
					fd.oo_class_assignment_data.applied_heuristic = {OO_FUNCTIONS_TRANSFORMER}.Heuristic_struct_pointer_arg
				then
					fd.accept (l_visitor)
				end
			end

			across l_update_statements as c loop
				insert_adjacent_statement (
						c.item,
						create {AST_EXPRESSION_STATEMENT}.make (create {AST_EIFFEL_EXPRESSION}.make_mixed (
								<<"update_class_fields", " -- Current might have been changed; synchronize">>,
								create {AST_BASIC_TYPE}.make (Ast_bt_void)
							)),
						True
					)
			end
		end

end
