class
	TYPEDEF_REMOVER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			end_visit_typedef_declaration,
			end_visit_typedef_type,
			end_visit_asm_operand,
			end_visit_array_type
		end

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			create typedefs.make (16)
			create structs.make (16)
			across transformer.tree.declarations as c loop
				if attached {AST_STRUCT_DECLARATION} c.item as tmp and then not tmp.is_forward_declaration then
					structs.extend (tmp, tmp.identifier)
				end
			end
			transformer.tree.accept (Current)
			typedefs := Void
			structs := Void
			Result := True
		end

	description: STRING
		do
			Result := "Removing typedefs"
		end

feature {ANY} -- AST_DEFAULT_VISITOR

	end_visit_typedef_declaration (node: AST_TYPEDEF_DECLARATION)
		do
			delete_node (node)
			typedefs[node.identifier] := node
			if
				attached {AST_STRUCT_TYPE} node.type as tmp and then
				structs.has (tmp.identifier) -- might be a struct without declaration (e.g. _XDisplay in xeyes)
			then
				structs[tmp.identifier].typedef_name := node.identifier
			end
		end

	end_visit_typedef_type (node: AST_TYPEDEF_TYPE)
		do
			replace_node (node, typedefs[node.identifier].type.ast_twin)
		end

	end_visit_asm_operand (node: AST_ASM_OPERAND)
		do
			-- also remove typedef types here
			node.orig_c_expression.accept (Current)
		end

	end_visit_array_type (node: AST_ARRAY_TYPE)
		do
			-- also remove typedef types here
			if node.orig_c_dimension /= Void then
				node.orig_c_dimension.accept (Current)
			end
		end

feature {NONE} -- temporary data

	typedefs: HASH_TABLE [AST_TYPEDEF_DECLARATION, STRING]
		-- Identifier (e.g. __off_t) -> AST_TYPEDEF_DECLARATION (typedef long __off_t)

	structs: HASH_TABLE [AST_STRUCT_DECLARATION, STRING]

end
