class
	AST_STATISTICS_VISITOR

inherit
	AST_DEFAULT_VISITOR
		redefine
			visit_assembler_statement,
			visit_expression_statement,
			visit_function_call,
			visit_if_statement,
			visit_jump_statement,
			visit_labeled_statement,
			visit_switch_statement,
			visit_while_statement,
			end_visit_if_statement,
			end_visit_switch_statement,
			end_visit_while_statement
		end

create
	default_create

feature -- AST_DEFAULT_VISITOR

	visit_assembler_statement (node: AST_ASSEMBLER_STATEMENT): BOOLEAN
		do
			process_statement
		end

	visit_expression_statement (node: AST_EXPRESSION_STATEMENT): BOOLEAN
		do
			process_statement
		end

	visit_function_call (node: AST_FUNCTION_CALL): BOOLEAN
		do
			process_statement
		end

	visit_if_statement (node: AST_IF_STATEMENT): BOOLEAN
		do
			process_statement

			if node.is_skip_block then
				skip_blocks := skip_blocks + 1
			end

			increase_nesting
			Result := True
		end

	visit_jump_statement (node: AST_JUMP_STATEMENT): BOOLEAN
		do
			jump_statements := jump_statements + 1
			process_statement
		end

	visit_labeled_statement (node: AST_LABELED_STATEMENT): BOOLEAN
		do
			Result := True
		end

	visit_switch_statement (node: AST_SWITCH_STATEMENT): BOOLEAN
		do
			process_statement

			increase_nesting
			Result := True
		end

	visit_while_statement (node: AST_WHILE_STATEMENT): BOOLEAN
		do
			process_statement

			increase_nesting
			Result := True
		end

	end_visit_if_statement (node: AST_IF_STATEMENT)
		do
			decrease_nesting
		end

	end_visit_switch_statement (node: AST_SWITCH_STATEMENT)
		do
			decrease_nesting
		end

	end_visit_while_statement (node: AST_WHILE_STATEMENT)
		do
			decrease_nesting
		end

feature -- Access

	statements: INTEGER

	maximal_nesting: INTEGER

	average_nesting: DOUBLE
		do
			Result := nesting_sum / statements
		end

	jump_statements: INTEGER

	skip_blocks: INTEGER

	to_csv_string: STRING
		local
			fd: FORMAT_DOUBLE
		do
			create fd.make (2, 2)

			create Result.make_from_string (statements.out + ", ")
			Result.append (maximal_nesting.out + ", ")
			Result.append (fd.formatted (average_nesting) + ", ")
			Result.append (jump_statements.out + ", ")
			Result.append (skip_blocks.out)
		end

feature {NONE} -- Implementation

	current_nesting: INTEGER

	nesting_sum: INTEGER_64

	increase_nesting
		do
			current_nesting := current_nesting + 1
			if current_nesting > maximal_nesting then
				maximal_nesting := current_nesting
			end
		end

	decrease_nesting
		do
			current_nesting := current_nesting - 1
		end

	process_statement
		do
			statements := statements + 1
			nesting_sum := nesting_sum + current_nesting
		end

end
