class
	INTEGER_CONSTANT_EXPRESSION_EVALUATOR

inherit
	ANTLR_AST_TYPE_CONSTANTS

create
	make

feature {NONE}

	make (a_expression: AST_EXPRESSION)
		require
			a_expression /= Void
		local
			l_tmp: like process
		do
			l_tmp := process (a_expression)
			is_integer := l_tmp.is_integer
			integer := l_tmp.i
			natural := l_tmp.n
		end

feature {ANY}

	is_integer: BOOLEAN
		-- is `integer' or `natural' set?

	integer: INTEGER_32

	natural: NATURAL_32

feature {NONE}

	process (a_expression: AST_EXPRESSION): TUPLE [is_integer: BOOLEAN; i: INTEGER_32; n: NATURAL_32]
		require
			a_expression /= Void
		local
			l_tmp1, l_tmp2: like process
			l_hex_converter: HEXADECIMAL_STRING_TO_INTEGER_CONVERTER
		do
			if attached {AST_DECIMAL_CONSTANT} a_expression as dc then
				inspect dc.type_binding.type
				when Ast_bt_int, Ast_bt_long then
					Result := [True, dc.value.to_integer_32, {NATURAL_32} 0]
				when Ast_bt_unsigned_int, Ast_bt_unsigned_long then
					Result := [False, 0, dc.value.to_natural_32]
				end
			elseif attached {AST_HEX_CONSTANT} a_expression as hc then
				create l_hex_converter.make
				l_hex_converter.set_leading_separators (" ")
				l_hex_converter.set_trailing_separators (" ")
				l_hex_converter.set_leading_separators_acceptable (True)
				l_hex_converter.set_trailing_separators_acceptable (True)

				inspect hc.type_binding.type
				when Ast_bt_int, Ast_bt_long then
					l_hex_converter.parse_string_with_type (hc.value, {NUMERIC_INFORMATION}.Type_integer_32)
					Result := [True, l_hex_converter.parsed_integer_32, {NATURAL_32} 0]
				when Ast_bt_unsigned_int, Ast_bt_unsigned_long then
					l_hex_converter.parse_string_with_type (hc.value, {NUMERIC_INFORMATION}.Type_natural_32)
					Result := [False, 0, l_hex_converter.parsed_natural_32]
				end
			elseif attached {AST_UNARY_EXPRESSION} a_expression as ue then
				l_tmp1 := process (ue.expression)
				if l_tmp1.is_integer then
					inspect ue.type
					when C_op_plus then  Result := l_tmp1
					when C_op_minus then Result := [True, -l_tmp1.i, {NATURAL_32} 0]
					end
				else
					inspect ue.type
					when C_op_plus then  Result := l_tmp1
					when C_op_minus then Result := [False, 0, (-(l_tmp1.n.to_integer_32)).to_natural_32]
					end
				end
			elseif attached {AST_GENERIC_EXPRESSION} a_expression as ge then
				l_tmp1 := process (ge.left)
				l_tmp2 := process (ge.right)
				check l_tmp1.is_integer = l_tmp2.is_integer end
				if l_tmp1.is_integer then
					inspect ge.type
					when C_op_plus then        Result := [True, l_tmp1.i + l_tmp2.i, {NATURAL_32} 0]
					when C_op_minus then       Result := [True, l_tmp1.i - l_tmp2.i, {NATURAL_32} 0]
					when C_op_bitand then      Result := [True, l_tmp1.i & l_tmp2.i, {NATURAL_32} 0]
					when C_op_bitor then       Result := [True, l_tmp1.i | l_tmp2.i, {NATURAL_32} 0]
					when C_op_exor then        Result := [True, l_tmp1.i.bit_xor (l_tmp2.i), {NATURAL_32} 0]
					when C_op_mult then        Result := [True, l_tmp1.i * l_tmp2.i, {NATURAL_32} 0]
					when C_op_div then         Result := [True, l_tmp1.i // l_tmp2.i, {NATURAL_32} 0]
					when C_op_modulo then      Result := [True, l_tmp1.i \\ l_tmp2.i, {NATURAL_32} 0]
					when C_op_left_shift then  Result := [True, l_tmp1.i |<< l_tmp2.i, {NATURAL_32} 0]
					when C_op_right_shift then Result := [True, l_tmp1.i |>> l_tmp2.i, {NATURAL_32} 0]
					end
				else
					inspect ge.type
					when C_op_plus then   Result := [False, 0, l_tmp1.n + l_tmp2.n]
					when C_op_minus then  Result := [False, 0, l_tmp1.n - l_tmp2.n]
					when C_op_bitand then Result := [False, 0, l_tmp1.n & l_tmp2.n]
					when C_op_bitor then  Result := [False, 0, l_tmp1.n | l_tmp2.n]
					when C_op_exor then   Result := [False, 0, l_tmp1.n.bit_xor (l_tmp2.n)]
					when C_op_mult then   Result := [False, 0, l_tmp1.n * l_tmp2.n]
					when C_op_div then    Result := [False, 0, l_tmp1.n // l_tmp2.n]
					when C_op_modulo then Result := [False, 0, l_tmp1.n \\ l_tmp2.n]
					end
				end
			elseif attached {AST_PARENTHESIZED_EXPRESSION} a_expression as pe then
				Result := process (pe.expression)
			elseif
				attached {AST_EIFFEL_EXPRESSION} a_expression as ee and then
				ee.mixed.count = 2 and then
				attached {AST_EXPRESSION} ee.mixed.first as e and then
				attached {STRING_8} ee.mixed.last as s
			then
				l_tmp1 := process (e)
				if l_tmp1.is_integer then
					    if s.is_equal (".to_integer_32") then Result := l_tmp1
					elseif s.is_equal (".to_natural_32") then Result := [False, 0, l_tmp1.i.to_natural_32]
					else
						(create {EXCEPTIONS}).raise ("Unsupported cast: " + a_expression.c_code)
					end
				else
					    if s.is_equal (".to_integer_32") then Result := [True, l_tmp1.n.to_integer_32, {NATURAL_32} 0]
					elseif s.is_equal (".to_natural_32") then Result := l_tmp1
					else
						(create {EXCEPTIONS}).raise ("Unsupported cast: " + a_expression.c_code)
					end
				end
			else
				(create {EXCEPTIONS}).raise ("Unsupported operation: " + a_expression.generator)
			end
		end

end
