class
	AST_VIEWER

inherit
	EV_TITLED_WINDOW
		redefine
			initialize,
			show
		end
	VISION2_GC_WORKAROUND

create
	make

feature {NONE}

	make
		do
			default_create
			protect_window_from_gc (Current)
		end

feature {ANY}

	tree: EV_TREE

	show
		do
			from
				tree.start
			until
				tree.after
			loop
				add_mouse_listener (tree.item)
				tree.forth
			end
			Precursor
		end

feature {NONE}

	initialize
		local
			main_container: EV_VERTICAL_BOX
		do
			Precursor {EV_TITLED_WINDOW}

			create main_container
			extend (main_container)

			create tree
			main_container.extend (tree)

			close_request_actions.extend (agent close_request)

			set_title ("TreeViewer")
			set_size (800, 600)
		end

	add_mouse_listener (tree_item: EV_TREE_NODE)
		do
			tree_item.set_configurable_target_menu_mode
			tree_item.set_configurable_target_menu_handler (agent menu_handler (?, ?, ?, ?, tree_item))
			from
				tree_item.start
			until
				tree_item.after
			loop
				add_mouse_listener (tree_item.item)
				tree_item.forth
			end
		end

	menu_handler (a_menu: EV_MENU; a_target_list: ARRAYED_LIST [EV_PND_TARGET_DATA]; a_source: EV_PICK_AND_DROPABLE; a_source_pebble: ANY; a_selected_item: EV_TREE_NODE)
		local
			l_mi_expand_all: EV_MENU_ITEM
			l_mi_show_source: EV_MENU_ITEM
			l_mi_show_type_binding: EV_MENU_ITEM
			l_mi_show_cfg: EV_MENU_ITEM
		do
			a_selected_item.enable_select

			create l_mi_expand_all.make_with_text_and_action ("Expand All", agent expand_all_requested)
			if not a_selected_item.is_expandable then
				l_mi_expand_all.disable_sensitive
			end
			a_menu.extend (l_mi_expand_all)

			create l_mi_show_source.make_with_text_and_action("Show Source", agent show_source_requested)
			if not attached {AST_NODE} a_selected_item.data then
				l_mi_show_source.disable_sensitive
			end
			a_menu.extend (l_mi_show_source)

			create l_mi_show_type_binding.make_with_text_and_action("Show Type Binding", agent show_type_binding_requested)
			if not attached {AST_EXPRESSION} a_selected_item.data then
				l_mi_show_type_binding.disable_sensitive
			end
			a_menu.extend (l_mi_show_type_binding)

			create l_mi_show_cfg.make_with_text_and_action("Show Control Flow Graph", agent show_cfg_requested)
			if not (attached {AST_FUNCTION_DECLARATION} a_selected_item.data as fd and then fd.cfg /= Void) then
				l_mi_show_cfg.disable_sensitive
			end
			a_menu.extend (l_mi_show_cfg)
		end

	expand_all_requested
		do
			expand_all (tree.selected_item)
		end

	show_source_requested
		local
			source_viewer: SOURCE_VIEWER
		do
			create source_viewer.make
			source_viewer.source_code.set_text (({AST_NODE} #? tree.selected_item.data).c_code)
			source_viewer.show
		end

	show_type_binding_requested
		local
			l_message_dialog: EV_INFORMATION_DIALOG
		do
			create l_message_dialog.make_with_text (({AST_EXPRESSION} #? tree.selected_item.data).type_binding.eiffel_code)
			l_message_dialog.set_title ("Type Binding")
			l_message_dialog.show_modal_to_window (Current)
		end

	show_cfg_requested
		do
			(create {CFG_VIEWER}.make ({AST_FUNCTION_DECLARATION} #? tree.selected_item.data)).show
		end

	expand_all (cur: EV_TREE_NODE)
		do
			if cur.is_expandable then
				cur.expand
				from
					cur.start
				until
					cur.after
				loop
					expand_all (cur.item)
					cur.forth
				end
			end
		end

	close_request
		do
			destroy
		end

end
